/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.repository.util;

import java.io.OutputStream;
import java.io.PrintStream;
import java.text.DateFormat;
import java.util.Date;


/**
 * For package debugging only, will not be included in distribution.
 */
public final class Debug
{
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName     = "$Name:  $";
    
    /**
     * Set this variable to true for debuging.
     */
    private static boolean ON = false;
    private static PrintStream writer = System.err;
    private static DateFormat dateFormatter = DateFormat.getDateTimeInstance();
    
    
    public static void setTraceOn()
    {
        ON = true;
    }
    
    public static void setTraceOn(OutputStream out)
    {
        setTraceOn();
        setOutputStream(out);
    }
    
    public synchronized static void setOutputStream(OutputStream out)
    {
        writer = new PrintStream(out);
    }
    
    public synchronized static OutputStream getOutputStream()
    {
        return writer;
    }
    
    public static void setTraceOff()
    {
        ON = false;
    }
    
    public static boolean traceIsOn()
    {
        return ON;
    }
    
    /**
     * Print the debug message when <code>ON</code> is set to true.
     */
    public synchronized static void println()
    {
        if (ON)
        {
            writer.println();
            writer.flush();
        }
    }
    
    public synchronized static void print(Object o1)
    {
        if (ON)
        {
            printBanner();
            writer.println(o1.toString());
            writer.println();
            writer.flush();
        }
    }
    public synchronized static void print(Object o1, Object o2)
    {
        if (ON)
        {
            printBanner();
            writer.println(o1.toString());
            writer.println(o2.toString());
            writer.println();
            writer.flush();
        }
    }
    
    public synchronized static void print(Object o1, Object o2, Object o3)
    {
        if (ON)
        {
            printBanner();
            writer.println(o1.toString());
            writer.println(o2.toString());
            writer.println(o3.toString());
            writer.println();
            writer.flush();
        }
    }
    
    public synchronized static void print(Object o1, Object o2, Object o3, Object o4)
    {
        if (ON)
        {
            printBanner();
            writer.println(o1.toString());
            writer.println(o2.toString());
            writer.println(o3.toString());
            writer.println(o4.toString());
            writer.println();
            writer.flush();
        }
    }
    
    public synchronized static void print(Object o1, Object o2, Object o3, Object o4, Object o5)
    {
        if (ON)
        {
            printBanner();
            writer.println(o1.toString());
            writer.println(o2.toString());
            writer.println(o3.toString());
            writer.println(o4.toString());
            writer.println(o5.toString());
            writer.println();
            writer.flush();
        }
    }
    
    private static void printBanner()
    {
        writer.print("===== ");
        writer.print(dateFormatter.format(new Date(System.currentTimeMillis())));
        writer.print(" === Thread[");
        writer.print(Thread.currentThread().getName());
        writer.println("] =====");
    }
}

