/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xquery.typing.QType;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;
import org.xquark.xquery.xdbc.XResultSetImpl;

public class ReconstructionVisitor extends DefaultParserVisitor {

	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";

	private ContentHandler contenthandler = null;
	private LexicalHandler lexicalhandler = null;
	private ErrorHandler errorhandler = null;

	private XResultSetImpl resultSet = null;
	private XDBCResultSetInterface xdbcResultSet = null;

	private AttributesImpl listAttributes = null;

	private Object[] lastIdentifiers = null;
	private Object[] currentIdentifiers = null;
	private Object[] nextIdentifiers = null;
	private boolean noNext = false;
	private boolean noReconstructor = false;
	private boolean hasidentifiers = false;

	private boolean hasPrefixMapping = false;
	private NamespaceContextStack declarations = null;

	private String nameSpace = null;
	private String prefixName = null;
	private String localName = null;

	private int size = -1;
	private ArrayList varList = null;
	private int lastSkolemIDsSize = -1;
	private ArrayList lastDependIDs = null;
	;

	private boolean inAttribute = false;
	private String inAttributeResult = null;

	//private boolean swap = false;
	private boolean root = true;
	//private boolean next = false;
	//private boolean withoutID = false;
	//private int cptPass = 0;
	private boolean noResultSet = false;
	private boolean noReInit = false;
	private ArrayList skolemIDs = null;

	public void reInit() {
		if (!noReInit)
			noNext = false;
		else
			noReInit = false;
		root = true;
		hasPrefixMapping = false;
		hasidentifiers = false;
	}

	public ReconstructionVisitor(XResultSetImpl resultSet, ArrayList varList) {
		this.resultSet = resultSet;
		this.xdbcResultSet = resultSet.getXDBCResultSet();
		this.varList = varList;
		size = varList.size();
		currentIdentifiers = new Object[size];
		nextIdentifiers = new Object[size];
		lastIdentifiers = new Object[size];
		listAttributes = new AttributesImpl();
	}

	private void setIdentifiers(boolean force) throws XQueryException {
		if (!force && hasidentifiers)
			return;
		try {
			if (xdbcResultSet.hasNext())
				xdbcResultSet.getNextIdentifiers(nextIdentifiers);
			if (!this.noResultSet)
				xdbcResultSet.getIdentifiers(currentIdentifiers);
			hasidentifiers = true;
		} catch (XMLDBCException xmldbce) {
			throw new XQueryException(xmldbce.getMessage());
		}
	}

	public void setContentHandler(ContentHandler handler) {
		if (handler == null)
			throw new NullPointerException("ContentHandler can not be null.");
		contenthandler = handler;
	}

	/**
	 * @see XDBCResultSetInterface#setLexicalHandler
	 */
	public void setLexicalHandler(LexicalHandler handler) {
		if (handler == null)
			throw new NullPointerException("LexicalHandler can not be null.");
		lexicalhandler = handler;
	}

	/**
	 * @see XDBCResultSetInterface#setErrorHandler
	 */
	public void setErrorHandler(ErrorHandler handler) {
		if (handler == null)
			throw new NullPointerException("ErrorHandler can not be null.");
		errorhandler = handler;
	}

	public boolean getOnGoingSkolemID() {
		if (currentIdentifiers != null)
			return true;
		else
			return false;
	}

	public boolean getNoReconstructor() {
		return noReconstructor;
	}

	public boolean getNoNext() {
		return noNext;
	}

	public void setNoNext(boolean noNext) {
		this.noNext = noNext;
	}

	public void setNoResultSet(boolean noResultSet) {
		this.noResultSet = noResultSet;
	}

	public void setNamespaceDecl(NamespaceContextStack declarations) {
		hasPrefixMapping = true;
		this.declarations = declarations;
	}

	// visitor ...

	//	public void visit(AggregateFunctionCall arg) throws XQueryException

	public void visit(AttributeValuePair arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:AttributeValuePair ( "+arg+" )");
		String nameSpaceAtt = null;
		// String prefixNameAtt = null;
		String localNameAtt = null;
		String value = null;

		XQueryExpression attrName = arg.getAttributeName();
		//org.xquark.xml.misc.Debug.print(this,"attrName: "+attrName.getClass());
		if (attrName instanceof QName) {
			QName qname = (QName) attrName;
			nameSpaceAtt = qname.getNameSpace();
			// prefixNameAtt = qname.getPrefixName();
			localNameAtt = qname.getLocalName();
		} else if (attrName instanceof Variable) {
			if (!noResultSet)
				localNameAtt = fetch(attrName.toString(), 0, attrName.getLoopIDs());
		} else if (attrName instanceof LocatedExpression) {
			inAttribute = true;
			attrName.accept(this);
			inAttribute = false;
			localNameAtt = inAttributeResult;
		} else if (attrName instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) attrName).getSubExpressions().size() == 1) {
			XQueryExpression tmpExpr = (XQueryExpression) ((XQueryExpressionSequence) attrName).getSubExpressions().get(0);
			inAttribute = true;
			tmpExpr.accept(this);
			inAttribute = false;
			localNameAtt = inAttributeResult;
			// localNameAtt = xdbcResultSet.fetch(tmpExpr.toString(),0,tmpExpr.getLoopIDs());
		} else {
			//org.xquark.xml.misc.Debug.print(this,"For now, just a QName, a Variable, a LocatedExpression or a XQueryExpressionSequence with a single XQueryExpression, check attribute name: "+attrName);
			return;
		}

		XQueryExpression attrValue = arg.getAttributeValue();
		//org.xquark.xml.misc.Debug.print(this,"attrValue: "+attrValue.getClass());
		if (attrValue instanceof ValueString)
			value = ((ValueString) attrValue).getStringValue();
		else if (attrValue instanceof Variable) {
			if (!noResultSet)
				value = fetch(attrValue.toString(), 0, attrValue.getLoopIDs());
		} else if (attrValue instanceof LocatedExpression) {
			inAttribute = true;
			attrValue.accept(this);
			inAttribute = false;
			value = inAttributeResult;
		} else if (attrValue instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) attrValue).getSubExpressions().size() == 1) {
			XQueryExpression tmpExpr = (XQueryExpression) ((XQueryExpressionSequence) attrValue).getSubExpressions().get(0);
			inAttribute = true;
			tmpExpr.accept(this);
			inAttribute = false;
			value = inAttributeResult;
			// value = xdbcResultSet.fetch(tmpExpr.toString(),0,tmpExpr.getLoopIDs());
		} else {
			//org.xquark.xml.misc.Debug.print(this,"For now, just a ValueString, a LocatedExpression or a XQueryExpressionSequence with a single XQueryExpression, check attribute value: "+attrValue);
			return;
		}

		// test if attribute already exists
		if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) != null)
			throw new ReconstructionException("An element cannot have several attributes with the same name.");
		// add attribute
		listAttributes.addAttribute((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt, "", "STRING", (value == null ? "" : value));
		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:AttributeValuePair ( "+arg+" )");
	}

	//	public void visit(BinOpANDExpression arg) throws XQueryException;
	//	public void visit(BinOpORExpression arg) throws XQueryException;
	//	public void visit(CastTreatExpression arg) throws XQueryException;

	public void visit(CData arg) throws XQueryException {
		try {
			if (lexicalhandler != null)
				lexicalhandler.startCDATA();
			characters(arg.getCData().toCharArray(), 0, arg.getCData().length());
			if (lexicalhandler != null)
				lexicalhandler.endCDATA();
		} catch (Exception e) {
			throw new XQueryException(e.getMessage());
		}
	}

	//	public void visit(ContextDeclaration arg) throws XQueryException;
	//	public void visit(Dereference arg) throws XQueryException;

	public void visit(Element arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( "+arg+" )");
		if (declarations != null)
			declarations.pushContext();
		HashMap localPrefixes = arg.getLocalPrefixes();
		if (localPrefixes != null) {
			Iterator it = localPrefixes.keySet().iterator();
			String prefix = null;
			while (it.hasNext()) {
				prefix = (String) it.next();
				declarations.declarePrefix(prefix, (String) localPrefixes.get(prefix));
			}
		}
		if (!hasPrefixMapping && localPrefixes != null)
			hasPrefixMapping = true;
		boolean isSkolemID = false;
		boolean isDependID = false;
		boolean putPrefixMapping = false;
		noReconstructor = true;
		ArrayList skolemIDs = arg.getSkolemIDs();
		this.skolemIDs = skolemIDs;
		if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
			isSkolemID = true;
		ArrayList dependIDs = arg.getDependIDs();
		if (dependIDs != null && !dependIDs.isEmpty())
			isDependID = true;
		boolean loop = arg.getLoop();

		//		org.xquark.xml.misc.Debug.print(this,"***************** skolemIDs: "+skolemIDs);
		//		org.xquark.xml.misc.Debug.print(this,"***************** isSkolemID: "+isSkolemID);
		//		org.xquark.xml.misc.Debug.print(this,"***************** dependIDs: "+dependIDs);
		//		org.xquark.xml.misc.Debug.print(this,"***************** isDependID: "+isDependID);
		//		org.xquark.xml.misc.Debug.print(this,"***************** loop: "+loop);

		XQueryExpression startTag = arg.getStartTag();
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element <"+startTag+">");

		// CASE : this element has no skolem ids
		if (!isSkolemID) {
			// CASE : the element is written until there are no more results available
			// ERASE  
			if (loop) {
				// no not continue if dependIDs changed
				if (this.noResultSet && dependIDs != lastDependIDs)
					return;
				while (true) {
					//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> <"+startTag+">");
					startTag.accept(this);
					// case of namespace in top element
					if (hasPrefixMapping) {
						startPrefixMapping(root);
						hasPrefixMapping = false;
						putPrefixMapping = true;
					}
					// prepare attributes
					prepareAttributes(arg);
					// case of attributes (to be evaluated in sons) to add at <start_element>
					int numAtt = buildAttributes(arg);
					// WRITE START
					purgeElements();
					startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
					// WRITE SONS
					ArrayList subExpressions = arg.getSubExpressions();
					if (subExpressions != null) {
						boolean previousRoot = root;
						root = false;
						for (int i = numAtt; i < subExpressions.size(); i++) {
							XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
							((XQueryExpression) subExpressions.get(i)).accept(this);
						}
						root = previousRoot;
					}
					// WRITE STOP
					QName endTag = (QName) arg.getEndTag();
					if (endTag != null)
						endTag.accept(this);
					else
						startTag.accept(this);
					purgeElements();
					endElement(nameSpace, localName, "");
					// case of namespace in top element
					if (putPrefixMapping) {
						hasPrefixMapping = true;
						putPrefixMapping = false;
						endPrefixMapping(root);
					}
					//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> </"+startTag+">");
					noReconstructor = true;
					if (hasNext())
						nextAsSAX();
					else
						break;
				}
			}
			// CASE : no loop and element has depend ids
			else if (isDependID) {
				noNext = false;
				boolean changeDependIDs = true;
				/// !!! DANGER !!! THIS is stupid code!!! ( )
				try {
					this.setIdentifiers(false);
					//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
				} catch (XQueryException e) {
					changeDependIDs = false;
				}
				while (!isEmpty(dependIDs) && changeDependIDs) {
					//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && DEPEND ID) -> <"+startTag+">");
					// case of namespace in top element
					startTag.accept(this);
					if (hasPrefixMapping) {
						startPrefixMapping(root);
						hasPrefixMapping = false;
						putPrefixMapping = true;
					}
					// prepare attributes
					prepareAttributes(arg);
					// case of attributes (to be evaluated in sons) to add at <start_element>
					int numAtt = buildAttributes(arg);
					// WRITE START
					purgeElements();
					startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
					// set lastSkolemIDsSize
					if (skolemIDs != null)
						lastSkolemIDsSize = skolemIDs.size();
					// WRITE SONS
					ArrayList subExpressions = arg.getSubExpressions();
					if (subExpressions != null) {
						boolean previousRoot = root;
						root = false;
						for (int i = numAtt; i < subExpressions.size(); i++) {
							XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
							((XQueryExpression) subExpressions.get(i)).accept(this);
						}
						root = previousRoot;
					}
					// WRITE STOP
					QName endTag = (QName) arg.getEndTag();
					if (endTag != null)
						endTag.accept(this);
					else
						startTag.accept(this);
					purgeElements();
					endElement(nameSpace, localName, "");
					// case of namespace in top element
					if (putPrefixMapping) {
						hasPrefixMapping = true;
						putPrefixMapping = false;
						endPrefixMapping(root);
					}
					//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && DEPEND ID) -> </"+startTag+">");
					//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
					// ERASE  
					/*
					xdbcResultSet.getIdentifiers(currentIdentifiers);
					 */
					if (!hasNext())
						break;
					// ERASE  
					/*
					xdbcResultSet.getNextIdentifiers(nextIdentifiers);
					 */
					//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
					//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
					changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
					if (!changeDependIDs) {
						noNext = false;
						break;
					} else if (!isEmpty(dependIDs))
						noNext = false;
					if (!noNext) {
						noReconstructor = true;
						//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
						if (!hasNext())
							break;
						identifierCopy(currentIdentifiers, lastIdentifiers);
						nextAsSAX();
						this.setIdentifiers(true);
						//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
						//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
					}
					if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
						noNext = true;
				}
			} else {
				//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITHOUT SKOLEM ID ) -> <"+startTag+">");
				startTag.accept(this);
				// case of namespace in top element
				if (hasPrefixMapping) {
					startPrefixMapping(root);
					hasPrefixMapping = false;
					putPrefixMapping = true;
				}
				// prepare attributes
				prepareAttributes(arg);
				// case of attributes (to be evaluated in sons) to add at <start_element>
				int numAtt = buildAttributes(arg);
				// WRITE START
				purgeElements();
				startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
				// WRITE SONS
				ArrayList subExpressions = arg.getSubExpressions();
				if (subExpressions != null) {
					boolean previousRoot = root;
					root = false;
					for (int i = numAtt; i < subExpressions.size(); i++) {
						XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
						((XQueryExpression) subExpressions.get(i)).accept(this);
					}
					root = previousRoot;
				}
				// WRITE STOP
				QName endTag = (QName) arg.getEndTag();
				if (endTag != null)
					endTag.accept(this);
				else
					startTag.accept(this);
				purgeElements();
				endElement(nameSpace, localName, "");
				// case of namespace in top element
				if (putPrefixMapping) {
					hasPrefixMapping = true;
					putPrefixMapping = false;
					endPrefixMapping(root);
				}
				//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITHOUT SKOLEM ID ) -> </"+startTag+">");
			}
		}
		// there is at least a skolemID
		else {
			//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITH SKOLEM ID ) -> <"+startTag+">");

			// set the current identifier and the next
			if (isSkolemID && !arg.getRoot()) {
				this.setIdentifiers(false);
				//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
			}
			//org.xquark.xml.misc.Debug.print(this,"***************** dependIDs: "+dependIDs);
			// there is no dependID
			if (dependIDs == null && !arg.getRoot()) {
				// CODE NOT USED !!! AS FAR AS I CAN TELL
				if (loop) {
					while (true) {
						//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> <"+startTag+">");
						startTag.accept(this);
						// case of namespace in top element
						if (hasPrefixMapping) {
							startPrefixMapping(root);
							hasPrefixMapping = false;
							putPrefixMapping = true;
						}
						// prepare attributes
						prepareAttributes(arg);
						// case of attributes (to be evaluated in sons) to add at <start_element>
						int numAtt = buildAttributes(arg);
						// WRITE START
						purgeElements();
						startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
						// set lastSkolemIDsSize
						if (skolemIDs != null)
							lastSkolemIDsSize = skolemIDs.size();
						// WRITE SONS
						ArrayList subExpressions = arg.getSubExpressions();
						if (subExpressions != null) {
							boolean previousRoot = root;
							root = false;
							for (int i = numAtt; i < subExpressions.size(); i++) {
								XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
								((XQueryExpression) subExpressions.get(i)).accept(this);
							}
							root = previousRoot;
						}

						// WRITE STOP
						QName endTag = (QName) arg.getEndTag();
						if (endTag != null)
							endTag.accept(this);
						else
							startTag.accept(this);
						purgeElements();
						endElement(nameSpace, localName, "");
						// case of namespace in top element
						if (putPrefixMapping) {
							hasPrefixMapping = true;
							putPrefixMapping = false;
							endPrefixMapping(root);
						}
						//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> </"+startTag+">");
						noReconstructor = true;
						noNext = false;
						if (hasNext())
							nextAsSAX();
						else
							break;
					}
				} else {
					//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITH SKOLEM ID && NO DEPEND ID ) -> <"+startTag+">");
					startTag.accept(this);
					// case of namespace in top element
					if (hasPrefixMapping) {
						startPrefixMapping(root);
						hasPrefixMapping = false;
						putPrefixMapping = true;
					}
					// prepare attributes
					prepareAttributes(arg);
					// case of attributes (to be evaluated in sons) to add at <start_element>
					int numAtt = buildAttributes(arg);
					// WRITE START
					purgeElements();
					startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
					// set lastSkolemIDsSize
					if (skolemIDs != null)
						lastSkolemIDsSize = skolemIDs.size();
					// WRITE SONS
					ArrayList subExpressions = arg.getSubExpressions();
					if (subExpressions != null) {
						boolean previousRoot = root;
						root = false;
						for (int i = numAtt; i < subExpressions.size(); i++) {
							XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
							((XQueryExpression) subExpressions.get(i)).accept(this);
						}
						root = previousRoot;
					}
					// WRITE STOP
					QName endTag = (QName) arg.getEndTag();
					if (endTag != null)
						endTag.accept(this);
					else
						startTag.accept(this);
					purgeElements();
					endElement(nameSpace, localName, "");
					// case of namespace in top element
					if (putPrefixMapping) {
						hasPrefixMapping = true;
						putPrefixMapping = false;
						endPrefixMapping(root);
					}
					//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITH SKOLEM ID && NO DEPEND ID ) -> </"+startTag+">");
				}
			} else // dependIDs != null or arg element is root
				{
				//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITH SKOLEM ID && WITH DEPEND ID ) -> <"+startTag+">");
				noNext = false;
				if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
					boolean changeDependIDs = true;
					//org.xquark.xml.misc.Debug.print(this,"***************** last < current (init) - isEmpty(dependIDs): "+isEmpty(dependIDs));
					//org.xquark.xml.misc.Debug.print(this,"***************** last < current (init) - changeDependIDs(dependIDs): "+changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs));

					boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
					// stop if the skolemIDs changed, and that dependIDs does not have an incidence
					if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
						if (declarations != null)
							declarations.popContext();
						return;
					}

					int firstDependIDPosition = 0;
					if (dependIDs != null && !dependIDs.isEmpty()) {
						firstDependIDPosition = getPosition((Variable) dependIDs.get(0)) + 1;
					}
					int skipEntries = -1;
					if (firstDependIDPosition > skolemIDs.size()) {
						skipEntries = lastSkolemIDsSize;
						while (!changeSkolemIDs && isEmpty(dependIDs)) {
							if (!hasNext())
								break;
							identifierCopy(currentIdentifiers, lastIdentifiers);
							nextAsSAX();
							this.setIdentifiers(true);
							//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
							//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
							changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
							changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
						}
					}
					while (!isEmpty(dependIDs) && changeDependIDs) {
						//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITH SKOLEM ID && DEPEND ID) last < current -> <"+startTag+">");
						startTag.accept(this);
						// case of namespace in top element
						if (hasPrefixMapping) {
							startPrefixMapping(root);
							hasPrefixMapping = false;
							putPrefixMapping = true;
						}
						// prepare attributes
						prepareAttributes(arg);
						// case of attributes (to be evaluated in sons) to add at <start_element>
						int numAtt = buildAttributes(arg);
						// WRITE START
						purgeElements();
						startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
						// set lastSkolemIDsSize
						if (skolemIDs != null)
							lastSkolemIDsSize = skolemIDs.size();
						// WRITE SONS
						ArrayList subExpressions = arg.getSubExpressions();
						if (subExpressions != null) {
							boolean previousRoot = root;
							root = false;
							for (int i = numAtt; i < subExpressions.size(); i++) {
								XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
								((XQueryExpression) subExpressions.get(i)).accept(this);
							}
							root = previousRoot;
						}
						// WRITE STOP
						QName endTag = (QName) arg.getEndTag();
						if (endTag != null)
							endTag.accept(this);
						else
							startTag.accept(this);
						purgeElements();
						endElement(nameSpace, localName, "");
						// case of namespace in top element
						if (putPrefixMapping) {
							hasPrefixMapping = true;
							putPrefixMapping = false;
							endPrefixMapping(root);
						}
						//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITH SKOLEM ID && DEPEND ID) last < current -> </"+startTag+">");
						//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
						// ERASE  
						/*
						xdbcResultSet.getIdentifiers(currentIdentifiers);
						 */
						if (!hasNext())
							break;
						// ERASE  
						/*
						xdbcResultSet.getNextIdentifiers(nextIdentifiers);
						 */
						//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
						//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
						if (skipEntries == -1) {
							changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, firstDependIDPosition - 1);
							if (!changeDependIDs) {
								noNext = false;
								break;
							} else if (!isEmpty(dependIDs))
								noNext = false;
						}
						// ----------------------------
						// ADDED  
						//-----------------------------
						else {
							changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries);
							changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, skipEntries);
							if (!changeSkolemIDs) {
								noNext = false;
								boolean noDependIDs = true;
								boolean makeBreak = false;
								noReInit = false;
								while (!changeSkolemIDs && noDependIDs) {
									noReconstructor = true;
									//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
									if (!hasNext()) {
										makeBreak = true;
										break;
									}
									identifierCopy(currentIdentifiers, lastIdentifiers);
									nextAsSAX();
									this.setIdentifiers(true);
									//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
									//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
									changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries);
									noDependIDs = isEmpty(dependIDs);
								}
								if (makeBreak)
									break;
								noNext = true;
								if (!noDependIDs)
									noReInit = true;
								changeDependIDs = true;
							} else if (!isEmpty(dependIDs))
								noNext = false;
						}
						// ----------------------------
						// END ADDED  
						//-----------------------------
						if (!noNext && !changeSkolemIDs) {
							noReconstructor = true;
							//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
							if (!hasNext())
								break;
							identifierCopy(currentIdentifiers, lastIdentifiers);
							nextAsSAX();
							this.setIdentifiers(true);
							//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
							//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
							//if (skipEntries == 0) changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
							//else changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs, skipEntries);
						}
						if (skipEntries == -1) {
							if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
								noNext = true;
						} else {
							if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
								noNext = false; // CHANGE   03/07/2002 was initially true
						}
					}
				} else {
					boolean changeSkolemIDs = false;
					while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
						//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Element ( WITH SKOLEM ID && DEPEND ID) last = current -> <"+startTag+">");
						startTag.accept(this);
						// case of namespace in top element
						if (hasPrefixMapping) {
							startPrefixMapping(root);
							hasPrefixMapping = false;
							putPrefixMapping = true;
						}
						// prepare attributes
						prepareAttributes(arg);
						// case of attributes (to be evaluated in sons) to add at <start_element>
						int numAtt = buildAttributes(arg);
						purgeElements();
						//org.xquark.xml.misc.Debug.print(this,"***************** nameSpace: "+nameSpace);
						//org.xquark.xml.misc.Debug.print(this,"***************** localName: "+localName);
						startElement((nameSpace == null) ? "" : nameSpace, localName, "", listAttributes);
						// set lastSkolemIDsSize
						if (skolemIDs != null)
							lastSkolemIDsSize = skolemIDs.size();
						// set last dependIDs
						lastDependIDs = dependIDs;
						// WRITE SONS
						ArrayList subExpressions = arg.getSubExpressions();
						if (subExpressions != null) {
							boolean previousRoot = root;
							root = false;
							byte num = 0;
							for (int i = numAtt; i < subExpressions.size(); i++) {
								if (!((XQueryExpression) subExpressions.get(i) instanceof Value))
									num++;
								if (num > 1)
									break;
							}
							for (int i = numAtt; i < subExpressions.size(); i++) {
								XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
								// CHANGED   10/04/2003
								if (num == 1 && arg.getRoot())
									expr.setRoot(true);
								//if (subExpressions.size() == 1 && arg.getRoot()) expr.setRoot(true);
								//((XQueryExpression) subExpressions.get(i)).accept(this);
								expr.accept(this);
							}
							root = previousRoot;
						}
						// WRITE STOP
						QName endTag = (QName) arg.getEndTag();
						//org.xquark.xml.misc.Debug.print(this,"***************** endTag: "+endTag);
						if (endTag != null)
							endTag.accept(this);
						else
							startTag.accept(this);
						purgeElements();
						//org.xquark.xml.misc.Debug.print(this,"***************** nameSpace: "+nameSpace);
						//org.xquark.xml.misc.Debug.print(this,"***************** localName: "+localName);
						endElement(nameSpace, localName, "");
						// case of namespace in top element
						if (putPrefixMapping) {
							hasPrefixMapping = true;
							putPrefixMapping = false;
							endPrefixMapping(root);
						}
						//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITH SKOLEM ID && DEPEND ID) last = current -> </"+startTag+">");
						// ERASE  
						/*
						if (!arg.getRoot()) xdbcResultSet.getIdentifiers(currentIdentifiers);
						 */
						if (!hasNext())
							break;
						// ERASE  
						/*
						if (!arg.getRoot()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
						 */
						//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
						//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
						//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
						//org.xquark.xml.misc.Debug.print(this,"***************** hasNext: "+resultSet.hasNext());

						if (!arg.getRoot())
							changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
						//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
						//org.xquark.xml.misc.Debug.print(this,"***************** dependIDs: "+dependIDs);
						//org.xquark.xml.misc.Debug.print(this,"***************** changeSkolemIDs: "+changeSkolemIDs);
						if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
							noReconstructor = true;
							//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
							if (!hasNext())
								break;
							identifierCopy(currentIdentifiers, lastIdentifiers);
							nextAsSAX();
							this.setIdentifiers(true);
						}
						if (changeSkolemIDs)
							noNext = true;
						if (arg.getRoot())
							break;
					}
					// 	while (!changeSkolemIDs);
				}
				// noNext = true;
			}
			//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( WITH SKOLEM ID ) -> </"+startTag+">");
		}
		if (declarations != null)
			declarations.popContext();
		noReconstructor = false;
		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Element ( "+arg+" )");
	}

	private void outputLocatedExpression(LocatedExpression arg, int nodeAccessor) throws XQueryException {
		boolean attribute = false;
		// this is hacking!!! implicit data in attribute
//		if (arg.getQType() != null && arg.getQType().getSubClass() == QType.ATTRIBUTE)
//			attribute = true;
		if (!noResultSet) {
			if (!inAttribute && !attribute) {
				generate(arg.print(), nodeAccessor, arg.getLoopIDs());
			} else if (inAttribute) {
				inAttributeResult = fetch(arg.print(), nodeAccessor, arg.getLoopIDs());
				//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:LocatedExpression ( inAttributeResult && LOOP ) : " + inAttributeResult);
				// if (tmpStr != null) handler.characters(tmpStr.toCharArray(), 0, tmpStr.length());
//			} else {
//				String attrValue = fetch(arg.print(), 0, arg.getLoopIDs());
//				if (attrValue != null)
//					characters(attrValue.toCharArray(), 0, attrValue.length());
			}
		}
	}

	public void visit(LocatedExpression arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:LocatedExpression ( "+arg+" )");

		boolean putPrefixMapping = false;

		if (hasPrefixMapping) {
			startPrefixMapping(true);
			hasPrefixMapping = false;
			putPrefixMapping = true;
		}

		// //org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:LocatedExpression : "+arg.getQType());
		// //org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:LocatedExpression : inAttribute : " + inAttribute);
		ArrayList steps = arg.getSteps();

		// //org.xquark.xml.misc.Debug.print(this,"### steps.get(steps.size() - 1): "+steps.get(steps.size() - 1));
		// //org.xquark.xml.misc.Debug.print(this,"### steps.get(steps.size() - 1): "+((Step)steps.get(steps.size() - 1)).getExpression().getClass());

		int position = steps.size() - 1;
		Step stepText = (Step) steps.get(position);
		XQueryExpression stepTextExpr = stepText.getExpression();
		boolean hasSpecialStep = false;
		//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:LocatedExpression : last step : " + stepText);
		// String tmpStr = null;
		int nodeAccessor = 0;
		// the TEXT stuff should be extended to other if necessary
		//if (stepText.getExpression() instanceof FunctionTEXT) steps.remove(position);
		if (stepTextExpr instanceof NodeType) {
			switch (((NodeType) stepTextExpr).getKind()) {
				case NodeKind.NODE :
					if (stepText.getAxis() == Axis.CHILD) {
						nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_NODE;
						steps.remove(position);
						hasSpecialStep = true;
					} else if (stepText.getAxis() == Axis.SELF) {
						nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_SELF_NODE;
						steps.remove(position);
						hasSpecialStep = true;
					}
					break;
				case NodeKind.TEXT :
					if (stepText.getAxis() == Axis.CHILD) {
						nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_TEXT;
						steps.remove(position);
						hasSpecialStep = true;
					} else if (stepText.getAxis() == Axis.SELF) {
						nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT;
						steps.remove(position);
						hasSpecialStep = true;
					}
					break;
			}
		}
		//        else if (stepTextExpr instanceof FunctionDATA) {
		//            nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_TEXT;
		//            steps.remove(position);
		//        }

		ArrayList dependIDs = arg.getDependIDs();
		//org.xquark.xml.misc.Debug.print(this,"***************** dependIDs: "+dependIDs);

		// there is no dependID
		if (!arg.getRoot()) {
			if (arg.getLoop()) {
				while (true) {
					outputLocatedExpression(arg, nodeAccessor);
					noReconstructor = true;
					if (hasNext())
						nextAsSAX();
					else
						break;
				}
			} else if (dependIDs != null && !dependIDs.isEmpty()) {
				noNext = false;
				boolean changeDependIDs = true;
				/// !!! DANGER !!! THIS is stupid code!!! ( )
				try {
					this.setIdentifiers(false);
					//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
				} catch (XQueryException e) {
					changeDependIDs = false;
				}
				while (!isEmpty(dependIDs) && changeDependIDs) {
					//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor WHILE ############################");

					outputLocatedExpression(arg, nodeAccessor);

					//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
					// ERASE  
					//xdbcResultSet.getIdentifiers(currentIdentifiers);
					if (!hasNext())
						break;
					// EARSE  
					//xdbcResultSet.getNextIdentifiers(nextIdentifiers);
					changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
					//org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs);
					if (!changeDependIDs) {
						noNext = true;
						break;
					}
					if (!noNext) {
						noReconstructor = true;
						//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
						if (!hasNext())
							break;
						if (!arg.getRoot())
							identifierCopy(currentIdentifiers, lastIdentifiers);
						nextAsSAX();
						this.setIdentifiers(true);
						changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
					}
					//org.xquark.xml.misc.Debug.print(this,"*****************  changeDependIDs: "+changeDependIDs);
				}
			} else {
				outputLocatedExpression(arg, nodeAccessor);
			}
		} else {
			noNext = false;
			boolean changeDependIDs = true;
			// set the current identiier and the next
			// xdbcResultSet.getIdentifiers(currentIdentifiers);
			//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
			//if (resultSet.hasNext()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
			//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
			if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** isEmpty(dependIDs): "+isEmpty(dependIDs));
				if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs));
					while (arg.getRoot() || changeDependIDs) {
						//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor WHILE ############################");
						outputLocatedExpression(arg, nodeAccessor);
						//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
						// EARSE  
						//if (!arg.getRoot()) xdbcResultSet.getIdentifiers(currentIdentifiers);
						if (!hasNext())
							break;
						// ERASE  
						//if (!arg.getRoot()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
						if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs));
							if (!arg.getRoot() && !changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs)) {
								noNext = true;
								break;
							}
						if (!noNext) {
							noReconstructor = true;
							//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
							if (!hasNext())
								break;
							if (!arg.getRoot())
								identifierCopy(currentIdentifiers, lastIdentifiers);
							nextAsSAX();
							// ERASE  
							//if (!arg.getRoot()) xdbcResultSet.getIdentifiers(currentIdentifiers);
							//if (resultSet.hasNext() && !arg.getRoot()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
							changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs);
						}
						//org.xquark.xml.misc.Debug.print(this,"*****************  changeDependIDs: "+changeDependIDs);
					}
		}
		//if (stepText.getExpression() instanceof FunctionTEXT) steps.add(stepText);
		if (hasSpecialStep)
			steps.add(stepText);
		//        else if (stepTextExpr instanceof FunctionDATA) steps.add(stepText);

		if (putPrefixMapping) {
			hasPrefixMapping = true;
			putPrefixMapping = false;
			endPrefixMapping(true);
		}

		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:LocatedExpression ( "+arg+" )");
	}

	public void visit(QName arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:QName ( "+arg+" )");
		nameSpace = arg.getNameSpace();
		prefixName = arg.getPrefixName();
		localName = arg.getLocalName();
		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:QName ( "+arg+" )");
	}

	public void visit(Value arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Value ( "+arg+" )");
		String text = arg.getValue();
		characters(text.toCharArray(), 0, text.length());
		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Value ( "+arg+" )");
	}

	private void outputVariable(Variable arg) throws XQueryException {
		boolean attribute = false;
		// this is hacking!!! implicit data in attribute
//		if (arg.getQType() != null && arg.getQType().getSubClass() == QType.ATTRIBUTE)
//			attribute = true;
		if (!noResultSet) {
			if (!inAttribute && !attribute) {
				int flag = 0;
				if (arg.getExpression() instanceof LocatedExpression) {
					ArrayList steps = ((LocatedExpression) arg.getExpression()).getSteps();
					Step lastStep = (Step) steps.get(steps.size() - 1);
					XQueryExpression lastStepExpr = lastStep.getExpression();
					if (lastStepExpr instanceof NodeType && ((NodeType) lastStepExpr).getKind() == NodeKind.TEXT)
						if (lastStep.getAxis() == Axis.CHILD)
							flag = XDBCResultSetInterface.NODE_ACCESSOR_TEXT;
						else if (lastStep.getAxis() == Axis.SELF)
							flag = XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT;
				}
				generate(arg.print(), flag, arg.getLoopIDs());
			} else if (inAttribute) {
				inAttributeResult = fetch(arg.print(), 0, arg.getLoopIDs());
//			} else {
//				String attrValue = fetch(arg.print(), 0, arg.getLoopIDs());
//				if (attrValue != null)
//					characters(attrValue.toCharArray(), 0, attrValue.length());
			}
		}
	}

	public void visit(Variable arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START ReconstructionVisitor:Variable ( "+arg+" )");
		boolean putPrefixMapping = false;

		if (hasPrefixMapping) {
			startPrefixMapping(true);
			hasPrefixMapping = false;
			putPrefixMapping = true;
		}

		ArrayList dependIDs = arg.getDependIDs();
		//org.xquark.xml.misc.Debug.print(this,"***************** dependIDs: "+dependIDs);

		if (!arg.getRoot()) {
			if (arg.getLoop()) {
				while (true) {
					//org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable ( NO DEPEND ID && LOOP ) -> "+arg);
					//org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable ( NO DEPEND ID && LOOP ) -> "+arg.getExpression());
					outputVariable(arg);
					//org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable ( NO DEPEND ID && LOOP ) -> "+arg);
					noReconstructor = true;
					if (hasNext())
						nextAsSAX();
					else
						break;
				}
			} else if (dependIDs != null && !dependIDs.isEmpty()) {
				noNext = false;
				boolean changeDependIDs = true;
				/// !!! DANGER !!! THIS is stupid code!!! ( )
				try {
					this.setIdentifiers(false);
					//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
				} catch (XQueryException e) {
					changeDependIDs = false;
				}
				while (!isEmpty(dependIDs) && changeDependIDs) {
					//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor WHILE ############################");
					outputVariable(arg);
					//					if (!noResultSet) {
					//						if (!inAttribute && !attribute) {
					//							int flag = 0;
					//							if (arg.getExpression() instanceof LocatedExpression) {
					//								ArrayList steps = ((LocatedExpression) arg.getExpression()).getSteps();
					//								XQueryExpression lastStepExpr = ((Step) steps.get(steps.size() - 1)).getExpression();
					//								if (lastStepExpr instanceof NodeType && ((NodeType) lastStepExpr).getKind() == NodeKind.TEXT)
					//									flag = 1;
					//							}
					//							generate(arg.print(), flag, arg.getLoopIDs());
					//							// generate(arg.print(),0,arg.getLoopIDs());
					//						} else if (inAttribute && attribute) {
					//							inAttributeResult = fetch(arg.print(), 0, arg.getLoopIDs());
					//							//   28/08/2003break;
					//						} else {
					//							attrValue = fetch(arg.print(), 0, arg.getLoopIDs());
					//							characters(attrValue.toCharArray(), 0, attrValue.length());
					//						}
					//					}
					this.setIdentifiers(false);
					//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
					// ERASE  
					/*
					xdbcResultSet.getIdentifiers(currentIdentifiers);
					 */
					if (!hasNext())
						break;
					// ERASE  
					/*
					xdbcResultSet.getNextIdentifiers(nextIdentifiers);
					 */
					changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
					//org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs);
					if (!changeDependIDs) {
						noNext = true;
						break;
					}
					if (!noNext) {
						noReconstructor = true;
						//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
						if (!hasNext())
							break;
						identifierCopy(currentIdentifiers, lastIdentifiers);
						nextAsSAX();
						this.setIdentifiers(true);
						changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs);
					}
					//org.xquark.xml.misc.Debug.print(this,"*****************  changeDependIDs: "+changeDependIDs);
				}
			} else {
				//org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable (NO DEPEND ID) -> "+arg);
				//org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable (NO DEPEND ID) -> "+arg.getExpression());
				outputVariable(arg); //org.xquark.xml.misc.Debug.print(this,"IN   ReconstructionVisitor:Variable (NO DEPEND ID) -> "+arg);
			}
		} else // dependIDs != null or it is a root
			{
			noNext = false;
			boolean changeDependIDs = true;
			// set the current identiier and the next
			// xdbcResultSet.getIdentifiers(currentIdentifiers);
			//org.xquark.xml.misc.Debug.print(this,"***************** currentIdentifiers: "+printArray(currentIdentifiers));
			//if (resultSet.hasNext()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
			//org.xquark.xml.misc.Debug.print(this,"***************** nextIdentifiers: "+printArray(nextIdentifiers));
			/*
			if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** isEmpty(dependIDs): "+isEmpty(dependIDs));
			if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs));
			 */
			while (arg.getRoot() || (!isEmpty(dependIDs) && changeDependIDs)) {
				//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor WHILE ############################");
				//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:Variable (DEPEND ID) -> "+arg);
				outputVariable(arg);
				//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:Variable (DEPEND ID) -> "+arg);
				//org.xquark.xml.misc.Debug.print(this,"***************** noNext: "+noNext);
				// ERASE  
				//if (!arg.getRoot()) xdbcResultSet.getIdentifiers(currentIdentifiers);
				if (!hasNext())
					break;
				// ERASE  
				//if (!arg.getRoot()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
				//if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** isEmpty(dependIDs): "+isEmpty(dependIDs));
				//if (!arg.getRoot()) //org.xquark.xml.misc.Debug.print(this,"***************** changeDependIDs(dependIDs): "+changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs));
				if (!arg.getRoot() && (isEmpty(dependIDs) || !changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs))) {
					noNext = true;
					break;
				}
				if (!noNext) {
					noReconstructor = true;
					//org.xquark.xml.misc.Debug.print(this,"***************** resultSet.hasNext(): "+resultSet.hasNext());
					if (!hasNext())
						break;
					if (!arg.getRoot())
						identifierCopy(currentIdentifiers, lastIdentifiers);
					nextAsSAX();
					// ERASE  
					/*
					if (!arg.getRoot()) xdbcResultSet.getIdentifiers(currentIdentifiers);
					if (resultSet.hasNext() && !arg.getRoot()) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
					 */
					if (!arg.getRoot())
						changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs);
				}
				//org.xquark.xml.misc.Debug.print(this,"*****************  changeDependIDs: "+changeDependIDs);
			}
		}

		if (putPrefixMapping) {
			hasPrefixMapping = true;
			putPrefixMapping = false;
			endPrefixMapping(true);
		}

		//org.xquark.xml.misc.Debug.print(this,"STOP  ReconstructionVisitor:Variable ( "+arg+" )");
	}

	public void visit(XMLComment arg) throws XQueryException {
		try {
			if (lexicalhandler != null)
				lexicalhandler.comment(arg.getComment().toCharArray(), 0, arg.getComment().length());
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	public void visit(XMLProcessingInstruction arg) throws XQueryException {
		try {
			if (contenthandler != null)
				contenthandler.processingInstruction(arg.getPITarget(), arg.getData());
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		ArrayList subExpressions = arg.getSubExpressions();
		if (subExpressions != null)
			for (int i = 0; i < subExpressions.size(); i++)
				 ((XQueryExpression) subExpressions.get(i)).accept(this);
	}

	//*************************************************************************************

	/*
	returns
	-1 if error
	0 if no change
	n for change in order n element
	 */
	private int identifierCompare(Object[] list1, Object[] list2) throws ReconstructionException {
		//	//org.xquark.xml.misc.Debug.print(this,"START identifierCompare:Element ( "+skolemSize+" ' "+list1+" ' "+list2+" )");
		//org.xquark.xml.misc.Debug.print(this,"START identifierCompare:Element ( "+list1+" ' "+list2+" )");
		if (list1 == null && list2 != null) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" ) --> " +list2.length);
			return list2.length;
		}
		if (list1 == null && list2 == null) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" )");
			//			return -1;
			throw new ReconstructionException("Incorrects argumments in method identifierCompare");
		}
		int len1 = list1.length;
		//		if (len1 > skolemSize) len1 = skolemSize;
		int len2 = list2.length;
		//		if (len2 > skolemSize) len2 = skolemSize;
		//org.xquark.xml.misc.Debug.print(this,"len1: "+len1);
		//org.xquark.xml.misc.Debug.print(this,"len2: "+len2);
		int i = 0;
		while (i < len1 && i < len2) {
			//org.xquark.xml.misc.Debug.print(this,"list1["+i+"]: "+list1[i]);
			//org.xquark.xml.misc.Debug.print(this,"list2["+i+"]: "+list2[i]);
			if (list1[i] == null || list2[i] == null) {
				//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" ) --> " +(i + 1));
				return i + 1;
			}
			if (!list1[i].equals(list2[i])) {
				//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" ) --> " +(i + 1));
				return i + 1;
			}
			i++;
		}
		if (len1 != len2) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" ) --> " +(i + 1));
			return i + 1;
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  identifierCompare:Element ( "+list1+" ' "+list2+" ) --> 0");
		return 0;
	}

	private boolean changeDependIDs(Object[] list1, Object[] list2, ArrayList dependIDs) throws ReconstructionException {
		return changeDependIDs(list1, list2, null, dependIDs, -1);
	}

	private boolean changeDependIDs(Object[] list1, Object[] list2, ArrayList skolemIDs, ArrayList dependIDs, int maxSkolemSize) throws ReconstructionException {
		//org.xquark.xml.misc.Debug.print(this,"START changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" )");
		if (list1 == null && list2 != null) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> true");
			return true;
		}
		if (list1 == null && list2 == null) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> error");
			//throw new ReconstructionException("Incorrects argumments in method changeDependIDs");
			//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> true");
			return true;
		}

		//org.xquark.xml.misc.Debug.print(this,"IN changeDependIDs::maxSkolemSize:"+maxSkolemSize);
		//org.xquark.xml.misc.Debug.print(this,"IN  changeDependIDs - BEFORE ( "+dependIDs+" )");
		if (skolemIDs != null) {
			int position = -1;
			for (int i = 0; i < maxSkolemSize; i++) {
				position = getPosition((Variable) skolemIDs.get(i));
				//org.xquark.xml.misc.Debug.print(this,"IN changeDependIDs::list1["+i+"]: "+list1[i]);
				//org.xquark.xml.misc.Debug.print(this,"IN changeDependIDs::list2["+i+"]: "+list2[i]);
				//if (list1[position] == null && list2[position] == null) continue;	// OLIVIER 2002/09/23
				if (list1[position] == null || list2[position] == null)
					return true;
				if (!list1[position].equals(list2[position])) {
					//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> false");
					return false;
				}
			}
		}
		//org.xquark.xml.misc.Debug.print(this,"IN  changeDependIDs - VERIF ( "+dependIDs+" )");
		for (int j = 0; j < dependIDs.size(); j++) {
			int position = getPosition((Variable) dependIDs.get(j));
			//org.xquark.xml.misc.Debug.print(this,"IN changeDependIDs::list1["+j+"]: "+list1[position]);
			//org.xquark.xml.misc.Debug.print(this,"IN changeDependIDs::list2["+j+"]: "+list2[position]);
			if (list1[position] == null || list2[position] == null)
				return true;
			if (!list1[position].equals(list2[position])) {
				//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> true");
				return true;
			}
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  changeDependIDs ( "+list1+" ' "+list2+" ' "+dependIDs+" ) -> false");
		return false;
	}

	private boolean changeSkolemIDs(Object[] list1, Object[] list2, ArrayList skolemIDs) throws ReconstructionException {
		return changeSkolemIDs(list1, list2, skolemIDs, skolemIDs.size());
	}

	private boolean changeSkolemIDs(Object[] list1, Object[] list2, ArrayList skolemIDs, int maxSkolemSize) throws ReconstructionException {
		//org.xquark.xml.misc.Debug.print(this,"START changeSkolemIDs ( "+list1+" ' "+list2+" ' "+skolemIDs+" )");
		int position = -1;
		for (int j = 0; j < maxSkolemSize; j++) {
			//org.xquark.xml.misc.Debug.print(this,"IN changeSkolemIDs::skolemIDs["+j+"]: "+(Variable) skolemIDs.get(j));
			position = getPosition((Variable) skolemIDs.get(j));
			//org.xquark.xml.misc.Debug.print(this,"IN changeSkolemIDs::list1["+j+"]: "+list1[position]);
			//org.xquark.xml.misc.Debug.print(this,"IN changeSkolemIDs::list2["+j+"]: "+list2[position]);
			if (list1[position] == null || list2[position] == null)
				return true;
			if (!list1[position].equals(list2[position])) {
				//org.xquark.xml.misc.Debug.print(this,"STOP  changeSkolemIDs ( "+list1+" ' "+list2+" ' "+skolemIDs+" ) -> true");
				return true;
			}
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  changeSkolemIDs ( "+list1+" ' "+list2+" ' "+skolemIDs+" ) -> false");
		return false;
	}

	private void identifierCopy(Object[] original, Object[] copy) {
		if (copy == null)
			copy = new Object[size];
		for (int i = 0; i < size; i++)
			copy[i] = original[i];
	}

	private boolean isEmpty(ArrayList dependIDs) throws ReconstructionException {
		// Olivier: 2002/10/11
		if (dependIDs == null)
			return true;
		//org.xquark.xml.misc.Debug.print(this,"START isEmpty ( "+dependIDs+" )");
		/*
		 for (int i = 0; i < dependIDs.size(); i++) {
						//org.xquark.xml.misc.Debug.print(this,"IN isEmpty::currentIdentifiers["+i+"]: "+currentIdentifiers[i]);
		}
		 */
		int countIsEmpty = 0;
		for (int i = 0; i < dependIDs.size(); i++) {
			//org.xquark.xml.misc.Debug.print(this,"IN isEmpty::currentIdentifiers["+getPosition((Variable) dependIDs.get(i))+"]: "+currentIdentifiers[getPosition((Variable) dependIDs.get(i))]);
			Object identifier = currentIdentifiers[getPosition((Variable) dependIDs.get(i))];
			// if (identifier == null
			//	|| (identifier instanceof ArrayList && ((ArrayList) identifier).size() == 1 && ((ArrayList) identifier).get(0) == null)
			//	|| (identifier instanceof String && ((String)identifier).equals("-1"))
			//	) countIsEmpty++;
			if (identifier == null || (identifier instanceof ArrayList && ((ArrayList) identifier).size() == 1 && ((ArrayList) identifier).get(0) == null))
				countIsEmpty++;
		}
		if (countIsEmpty == dependIDs.size()) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  isEmpty ( "+dependIDs+" ) -> true");
			return true;
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  isEmpty ( "+dependIDs+" ) -> false");
		return false;
	}

	private int getPosition(Variable variable) throws ReconstructionException {
		int position = varList.indexOf(variable);
		if (position == -1)
			throw new ReconstructionException("This variable " + variable + " is not defined in the list of variables.");
		return position;
	}

	private void purgeElements() {
		if (nameSpace == null)
			nameSpace = "";
		if (prefixName == null)
			prefixName = "";
		if (localName == null)
			localName = "";
	}

	private String fetch(String path, int nodeAccessor, String loopid) throws XQueryException {
		try {
			return xdbcResultSet.fetch(path, nodeAccessor, loopid);
		} catch (XMLDBCException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void generate(String path, int nodeAccessor, String loopid) throws XQueryException {
		try {
			xdbcResultSet.generate(path, nodeAccessor, loopid);
		} catch (XMLDBCException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void characters(char[] ch, int start, int length) throws XQueryException {
		try {
			if (contenthandler != null)
				contenthandler.characters(ch, start, length);
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void startElement(String str1, String str2, String str3, AttributesImpl atts) throws XQueryException {
		try {
			if (contenthandler != null)
				contenthandler.startElement(str1, str2, str3, atts);
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void endElement(String str1, String str2, String str3) throws XQueryException {
		try {
			if (contenthandler != null)
				contenthandler.endElement(str1, str2, str3);
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private boolean hasNext() throws XQueryException {
		try {
			return resultSet.hasNext();
		} catch (XMLDBCException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void nextAsSAX() throws XQueryException {
		try {
			resultSet.nextAsSAX();
		} catch (Exception e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void startPrefixMapping(boolean first) throws XQueryException {
		try {
			if (declarations != null) {
				String prefix = null;
				ArrayList currentList = null;
				if (first)
					currentList = new ArrayList(declarations.getPrefixes());
				else
					currentList = new ArrayList(declarations.getDeclaredPrefixes());
				for (int i = 0; i < currentList.size(); i++) {
					prefix = (String) currentList.get(i);
					//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> startPrefixMapping : " + prefix + " , " + declarations.getNamespaceURI(prefix));
					if (contenthandler != null)
						contenthandler.startPrefixMapping(prefix, declarations.getNamespaceURI(prefix));
				}
			}
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void endPrefixMapping(boolean first) throws XQueryException {
		try {
			if (declarations != null) {
				String prefix = null;
				ArrayList currentList = null;
				if (first)
					currentList = new ArrayList(declarations.getPrefixes());
				else
					currentList = new ArrayList(declarations.getDeclaredPrefixes());
				for (int i = 0; i < currentList.size(); i++) {
					//org.xquark.xml.misc.Debug.print(this,"IN    ReconstructionVisitor:Element ( WITHOUT SKOLEM ID && WITH DEPEND ID (NO IDENTIFIER) ) -> endPrefixMapping : " + currentList.get(i));
					if (contenthandler != null)
						contenthandler.endPrefixMapping((String) currentList.get(i));
				}
			}
		} catch (SAXException e) {
			throw new XQueryException(e.getMessage());
		}
	}

	private void prepareAttributes(Element elt) throws XQueryException {
		listAttributes.clear();
		ArrayList attributes = elt.getAttributes();
		if (attributes != null) {
			for (int i = 0; i < attributes.size(); i++)
				 ((AttributeValuePair) attributes.get(i)).accept(this);
		}
	}

	private int buildAttributes(Element elt) throws XQueryException {
		ArrayList sons = elt.getSubExpressions();
		if (sons == null)
			return 0;
		int i = 0;
		for (; i < sons.size(); i++) {
			XQueryExpression son = (XQueryExpression) sons.get(i);
			QType qtype = son.getQType();
			if (!qtype.isAttribute())
				return i;
			String nameSpaceAtt = null;
			String localNameAtt = null;
			String value = null;
			ArrayList dependIDs = son.getDependIDs();
			if (dependIDs == null || dependIDs.isEmpty()) {
				if (qtype.isMultiple()) {
					throw new ReconstructionException("An element cannot have several attributes with the same name.");					
				}				
				// get nameSpaceAtt, localNameAtt and value from expression
				nameSpaceAtt = ((QName) qtype.getName()).getNameSpace();
				localNameAtt = ((QName) qtype.getName()).getLocalName();
				if (son instanceof Variable) {
					XQueryExpression expr = ((Variable) son).getExpression();
					if (!noResultSet)
						value = fetch(son.print(), 0, son.getLoopIDs());
				} else if (son instanceof LocatedExpression) {
					inAttribute = true;
					son.accept(this);
					inAttribute = false;
					value = inAttributeResult;
				} else if (son instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) son).getSubExpressions().size() == 1) {
					inAttribute = true;
					XQueryExpression firstSon = (XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0);
					firstSon.accept(this);
					inAttribute = false;
					// DANGER!!! Olivier: 2002/12/17
					if (firstSon instanceof AttributeValuePair && ((AttributeValuePair) firstSon).isComputed())
						continue;
					value = inAttributeResult;
				} else
					continue;
				// test if attribute already exists
				if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) != null)
					throw new ReconstructionException("An element cannot have several attributes with the same name.");
				// generate attribute
				listAttributes.addAttribute((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt, "", "STRING", (value == null ? "" : value));
			} else {
				//				qtype = qtype.getQType();
				
				boolean root = son.getRoot();
				if (son instanceof XQueryExpressionSequence) {
					dependIDs = ((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).getDependIDs();
					root = ((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).getRoot();
				}
				noNext = false;
				boolean changeDependIDs = true;
				this.setIdentifiers(false);
				/*
						try {
							// ERASE  
							xdbcResultSet.getIdentifiers(currentIdentifiers);
						}
						catch (XQueryException e) {
							changeDependIDs = false;
						}
				 */
				while (!isEmpty(dependIDs) && changeDependIDs) {

					// get nameSpaceAtt, localNameAtt and value from expression
					nameSpaceAtt = ((QName) qtype.getName()).getNameSpace();
					localNameAtt = ((QName) qtype.getName()).getLocalName();
					if (son instanceof Variable) {
						XQueryExpression expr = ((Variable) son).getExpression();
						if (!noResultSet)
							value = fetch(son.print(), 0, son.getLoopIDs());
					} else if (son instanceof LocatedExpression) {
						inAttribute = true;
						son.accept(this);
						inAttribute = false;
						value = inAttributeResult;
					} else if (son instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) son).getSubExpressions().size() == 1) {
						inAttribute = true;
						((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).accept(this);
						inAttribute = false;
						value = inAttributeResult;
					} else
						continue;
					// test if attribute already exists
					if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) != null)
						throw new ReconstructionException("An element cannot have several attributes with the same name.");
					// generate attribute
					listAttributes.addAttribute((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt, "", "STRING", (value == null ? "" : value));

					this.setIdentifiers(false);
					// ERASE  
					/*
					xdbcResultSet.getIdentifiers(currentIdentifiers);
					 */
					if (!hasNext())
						break;
					// ERASE  
					/*
					xdbcResultSet.getNextIdentifiers(nextIdentifiers);
					 */
					changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
					if (!changeDependIDs) {
						noNext = true;
						break;
					}
					if (!noNext) {
						noReconstructor = true;
						if (!hasNext())
							break;
						if (!root)
							identifierCopy(currentIdentifiers, lastIdentifiers);
						nextAsSAX();
						this.setIdentifiers(true);
						// ERASE  
						/*
						if (!root) xdbcResultSet.getIdentifiers(currentIdentifiers);
						if (resultSet.hasNext() && !root) xdbcResultSet.getNextIdentifiers(nextIdentifiers);
						 */
						changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs);
					}
				}
			}
		}
		return i;
	}

	// stuff for debug
	private String printArray(Object[] array) {
		StringBuffer buf = new StringBuffer();
		if (array != null)
			for (int i = 0; i < array.length; i++) {
				Object obj = array[i];
				if (obj == null)
					buf.append(" null ");
				else
					buf.append(array[i].toString() + " ");
			}
		return buf.toString();
	}
}
